# -*- coding: utf-8 -*-
"""
@author: P.Vana
"""
import sys
import os
import numpy as np
import math
import time as tm

import helper as hlp

##################################################
# Neuron: The main solver
##################################################
class Neuron:
	def __init__(self):
		# neuron position
		self.x = 0
		self.y = 0
		# associated goal position
		self.gx = None
		self.gy = None
		# index of the goal which inhibited this neuron
		self.goal_idx = None
		pass

	def adapt_neuron(self, beta, goal):
		self.x += beta * (goal[0] - self.x)
		self.y += beta * (goal[1] - self.y)
		pass

	def position(self):
		return ((self.x, self.y))
	
	def goal_position(self):
		return ((self.gx, self.gy))
	
	def update_goal_position(self, goal, radius):
		# TODO - replace this code by the altergnate goal position
		(self.gx, self.gy) = goal
		
		return ((self.gx, self.gy))

##################################################
# SOM_solver: The main solver
##################################################
class SOM_solver:
	def __init__(self, goals, radius):
		self.goals = goals # vector containint coordinates of N goal cities
		self.radius = radius # size of the goal neighborhoods
		
		N = len(goals) # number of goals
		M = int(2.5 * N) # number of neurons
		
		# set the learning rate based on the neuron count N (original version)
		self.sigma = 12.41 * N + 0.6
		# faster version -- used only for UIR
		self.sigma = N
		self.mi = 0.5 # learning rate 
		self.alpha = 0.1 # gain decreasing rate

		# compute center and deviations of the goals
		np_goals = np.array([np.array(g) for g in goals])
		center = np.average(np_goals, axis=0)
		dev = np.std(np_goals, axis=0)
		
		# prepare matrix of the M neurons
		self.neurons = [Neuron() for i in range(0,M)]
		# initialize neurons by an ellipse
		for i in range(0, M):
			angle = (2 * math.pi * i) / M 
			direction = np.multiply([math.cos(angle), math.sin(angle)], dev)
			pos = center + direction
			self.neurons[i].x = pos[0]
			self.neurons[i].y = pos[1]
		
		# stopping criteria
		self.max_iterations = 120
		self.max_error = max(dev) * 1e-5
		pass	

	def select_winner(self, goal_idx):
		(act_goal_x, act_goal_y) = self.goals[goal_idx]

		best_idx = None
		best_dst = float("inf")

		idx = 0
		for n in self.neurons:
			if n.goal_idx == None:
				# use squared distance for a faster computation
				dx = act_goal_x - n.x
				dy = act_goal_y - n.y
				dst = dx*dx + dy*dy
				if dst < best_dst:
					best_dst = dst
					best_idx = idx
			idx = idx + 1
	
		self.neurons[best_idx].goal_idx = goal_idx		
		return best_idx

	def neighborhood_fce(self, distance):
		return math.exp(-distance * distance / self.sigma / self.sigma)
	
	def run_som(self, visualize = False):
		# show the initial ring and wait
		if visualize:
			self.show_current_ring()
			hlp.pause(1)

		N = len(self.goals)
		M = len(self.neurons)
		
		distances = range(int(-0.2*M)+1, int(0.2*M))

		# start of the main cyclus	
		for learning_epoch in range(0,self.max_iterations):
			# clear values from the previous learning_epoch
			for n in self.neurons:
				n.goal_idx = self.gx = self.gy = None

			# chooce random order of the goals
			order = np.random.permutation(len(self.goals))

			# precompute betas for faster computation
			betas = [self.mi*self.neighborhood_fce(dst) for dst in distances]

			# choose the closest neuron for each goal
			for goal_idx in order:
				act_goal = self.goals[goal_idx]
				winner_idx = self.select_winner(goal_idx)
				winner = self.neurons[winner_idx]
				alternate_goal = winner.update_goal_position(act_goal, self.radius)

				for d_idx, distance in enumerate(distances):
					beta = betas[d_idx]
					adapt_idx = (winner_idx + distance + M) % M
					self.neurons[adapt_idx].adapt_neuron(beta, alternate_goal)

			# update learing parameters
			self.sigma = (1 - self.alpha) * self.sigma	

			error = self.statistics(learning_epoch)
			if error < self.max_error:
				break

			# show current ring
			if visualize:
				self.show_current_ring()
				#hlp.save_figure('img/iter-{:04d}.png'.format(learning_epoch))
				hlp.pause(0.05)

		return self.reconstruct_path()

	def reconstruct_path(self):
		inhibited_neurons = filter(lambda x: x.goal_idx != None, self.neurons)
		path = [n.goal_position() for n in inhibited_neurons]
		if len(path) > 0: path = path + [path[0]]
		return path

	# save the best solution, print stats and return current error
	def statistics(self, iteration):
		error_sq = 0
		for n in filter(lambda x : x.goal_idx != None, self.neurons):
			dx = n.x - n.gx
			dy = n.y - n.gy
			dst_sq = dx*dx + dy*dy
			error_sq = max(error_sq, dst_sq)
		error = math.sqrt(error_sq)

		print('iteration {}: max_error = {}, time = {} s'.format(iteration, error, tm.clock()))
		return error
	
### Helper functions only for visualization (SOM_solver) ###

	def show_ring(self, ring, color):
		hlp.clear()
		# show goals of the given TSP instance
		for g in self.goals:
			hlp.plot_circle(g, self.radius)
			hlp.plot_points([g], 'ro')
		inhibited_neurons = filter(lambda x: x.goal_idx != None, self.neurons)
		for n in inhibited_neurons:
			hlp.plot_points([n.goal_position()], 'rx')
		# show the final ring of the neurons (black)
		hlp.plot_points(ring, color)
		# highlight inhibited neurons with their associated goals (green)
		inhibited_neurons = filter(lambda x: x.goal_idx != None, self.neurons)
		for n in inhibited_neurons:
			hlp.plot_points([n.position(), self.goals[n.goal_idx]], 'g--')
			hlp.plot_points([n.position()], 'go')
		pass

	def show_current_ring(self):
		self.show_ring(self.get_neurons_path(), 'kx-')
		pass

	def show_final_path(self):
		self.show_ring(self.reconstruct_path(), 'rx-')
		pass

	def get_neurons_path(self):
		path = [(n.x, n.y) for n in self.neurons]
		if len(path) > 0: path = path + [path[0]]
		return path
	